﻿using System;

namespace COMET2Lib
{
    public class Interpreter
    {
        // 3つのフラグのマスク値
        private enum FlagMasks { OF = 4, SF = 2, ZF = 1 };

        // COMET IIの世界
        public ushort[] memory = new ushort[65536];// メモリ世界
        private ushort[] GR = new ushort[8];// レジスタ世界
        private ushort SP, FR;// レジスタ世界
        public ushort PR;// レジスタ世界

        // トレースモードならtrue
        public bool traceMode = false;

        // スーパバイザコールのINPUT処理
        private void svcInput()
        {
            string buffer = Console.ReadLine();
            GR[2] = 0xffff;
            if (buffer == null) return;
            GR[2] = (ushort)buffer.Length;
            for (int i = 0; i < buffer.Length; i++)
            {
                memory[GR[1] + i] = buffer[i];
            }
        }

        // スーパバイザコールのOUTPUT処理
        private void svcOutput()
        {
            int i;
            for (i = 0; i < GR[2]; i++)
                Console.Write((char)(memory[GR[1] + i] & 0xff));
            Console.WriteLine();
        }

        // フラグレジスタの設定
        private ushort setFR(ushort result, bool OFvalue)
        {
            FR = (ushort)(((result == 0) ? (int)FlagMasks.ZF : 0) | ((result & 0x8000) >> 14) | (OFvalue ? (int)FlagMasks.OF : 0));
            return result;
        }

        // 算術命令(……A)の場合のフラグレジスタの設定
        private ushort setFRbyA(int result)
        {
            return setFR((ushort)result, result < -32768 || result > 32767);
        }

        // 論理命令(……L)の場合のフラグレジスタの設定
        private ushort setFRbyL(int result)
        {
            return setFR((ushort)result, result < 0 || result > 65535);
        }

        // 仮想CPUインタプリタのメイン
        public void interprete()
        {
            if (traceMode) Console.WriteLine("starting machine codes");

            FR = 0; SP = 0xfffe;
            memory[0xfffe] = 0xf000;	// SVC 0=停止を書き込んでおく
            memory[0xffff] = 0x0000;
            memory[--SP] = 0xfffe;	//停止命令のあるアドレスをPUSHしておく
            for (; ; )
            {
                ushort firstOP = memory[PR++];
                ushort OP = (ushort)(firstOP >> 8);
                ushort mainOP = (ushort)(OP >> 4);
                ushort subOP = (ushort)(OP & 0xf);
                ushort r1 = (ushort)((firstOP >> 4) & 0xf);
                ushort r2 = (ushort)((firstOP & 0xf)), value = 0, i;

                if (traceMode)
                {
                    Console.WriteLine("{0:x4} {1:x4}", PR - 1, memory[PR - 1]);
                    for (i = 0; i < 8; i++) Console.Write("GR{0}={1} ", i, GR[i]);
                    Console.WriteLine("SP={0:x4} ZF={1} SF={2} OF={3}", SP, (FR & (ushort)FlagMasks.ZF) != 0, (FR & (ushort)FlagMasks.SF) != 0, (FR & (ushort)FlagMasks.OF) != 0);
                }

                // 分類1
                if ((mainOP >= 1 && mainOP <= 5 && subOP <= 3)
                    || mainOP == 6 || OP == 0x70 || OP == 0x80 || OP == 0xf0)
                {
                    value = (ushort)((r2 == 0 ? (ushort)0 : GR[r2]) + memory[PR++]);
                    if (OP == 0x10 || (mainOP >= 2 && mainOP <= 5))
                        value = memory[value];
                }
                else if (mainOP >= 1 && mainOP <= 5 && subOP >= 4)
                    value = (r2 == 0 ? (ushort)0 : GR[r2]);

                // 分類2
                switch (OP & 0xfb)
                {
                    case 0x10:    // LD
                        GR[r1] = setFR(value, false);
                        break;
                    case 0x20:	// ADDA
                        GR[r1] = setFRbyA((short)GR[r1] + (short)value);
                        break;
                    case 0x21:	// SUBA
                        GR[r1] = setFRbyA((short)GR[r1] - (short)value);
                        break;
                    case 0x22:	// ADDL
                        GR[r1] = setFRbyL(GR[r1] + value);
                        break;
                    case 0x23:	// SUBL
                        GR[r1] = setFRbyL(GR[r1] - value);
                        break;
                    case 0x30:	// AND
                        GR[r1] = setFR((ushort)(GR[r1] & value), false);
                        break;
                    case 0x31:	// OR
                        GR[r1] = setFR((ushort)(GR[r1] | value), false);
                        break;
                    case 0x32:	// XOR
                        GR[r1] = setFR((ushort)(GR[r1] ^ value), false);
                        break;
                    case 0x40:	// CPA
                        FR = (ushort)((GR[r1] == value ? 1 : 0) | ((short)GR[r1] < (short)value ? 2 : 0));
                        break;
                    case 0x41:	// CPL
                        FR = (ushort)((GR[r1] == value ? 1 : 0) | (GR[r1] < value ? 2 : 0));
                        break;
                    case 0x50:	// SLA
                        GR[r1] = setFR((ushort)(((GR[r1] << value) & 0x7fff) | (GR[r1] & 0x8000)), ((GR[r1] << value) & 0x8000) != 0);
                        break;
                    case 0x51:	// SRA
                        // 処理系によっては意図しない結果になる可能性がある
                        GR[r1] = setFR((ushort)((short)GR[r1] >> value), value > 0 && ((GR[r1] >> (value - 1)) & 1) != 0);
                        break;
                    case 0x52:	// SLL
                        GR[r1] = setFR((ushort)(GR[r1] << value), value > 0 && ((GR[r1] << (value - 1)) & 0x8000) != 0);
                        break;
                    case 0x53:	// SRL
                        GR[r1] = setFR((ushort)(GR[r1] >> value), value > 0 && ((GR[r1] >> (value - 1)) & 1) != 0);
                        break;
                    default:
                        switch (OP)
                        {
                            case 0x00:	// NOP
                                break;
                            case 0x11:    // ST r,adr,x store
                                memory[value] = GR[r1];
                                break;
                            case 0x12:    // LAD r,adr,x load address
                                GR[r1] = value;
                                break;
                            case 0x61:	// JMI
                                if ((FR & (ushort)FlagMasks.SF) != 0) PR = value;
                                break;
                            case 0x62:	// JNZ
                                if ((FR & (ushort)FlagMasks.ZF) == 0) PR = value;
                                break;
                            case 0x63:	// JZE
                                if ((FR & (ushort)FlagMasks.ZF) != 0) PR = value;
                                break;
                            case 0x64:	// JUMP
                                PR = value;
                                break;
                            case 0x65:	// JPL
                                if ((FR & ((ushort)FlagMasks.SF | (ushort)FlagMasks.ZF)) == 0) PR = value;
                                break;
                            case 0x66:	// JOV
                                if ((FR & (ushort)FlagMasks.OF) != 0) PR = value;
                                break;
                            case 0x70:	// PUSH
                                memory[--SP] = value;
                                break;
                            case 0x71:	// POP
                                GR[r1] = memory[SP++];
                                break;
                            case 0x80:	// CALL
                                memory[--SP] = PR;
                                PR = value;
                                break;
                            case 0x81:	// RET
                                PR = memory[SP++];
                                break;
                            case 0xf0:	// SVC
                                switch (value)
                                {
                                    case 0:	// EXIT
                                        return;
                                    case 1:	// IN
                                        svcInput();
                                        break;
                                    case 2:	// OUT
                                        svcOutput();
                                        break;
                                }
                                break;
                        }
                        break;
                }
            }
        }
    }
}
